CREATE OR REPLACE FUNCTION upsert_row(_t regclass, rowId UUID, d JSONB) RETURNS INT AS $$
DECLARE
	affectedRow int;
BEGIN
    LOOP
        -- First try to update the key
        EXECUTE 'UPDATE '
            ||_t::regclass
            ||' SET data = $1 WHERE _id = $2'
            USING d, rowId;

		-- Note in particular that EXECUTE changes the output of GET DIAGNOSTICS, but does not change FOUND.
		-- So don't use 'IF found THEN' to tell if any row is updated
		GET DIAGNOSTICS affectedRow = ROW_COUNT;
		IF affectedRow = 1 THEN
            RETURN 1;
        END IF;

        -- not there, so try to insert the key
        -- if someone else inserts the same key concurrently,
        -- we could get a unique-key failure
        BEGIN
            EXECUTE 'INSERT INTO '
                || _t::regclass
                ||' (_id, data) VALUES ($1, $2)'
                USING rowId, d;
            GET DIAGNOSTICS affectedRow = ROW_COUNT;
            RETURN affectedRow;
        EXCEPTION WHEN unique_violation THEN
        -- Do nothing, and loop to try the UPDATE again.
        END;
    END LOOP;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION insert_row(_t regclass, rowId UUID, d JSONB) RETURNS JSONB AS $$
declare
	affectedRow int;
BEGIN
    EXECUTE 'INSERT INTO '
        || _t::regclass
        ||' (_id, data) VALUES ($1, $2)'
        USING rowId, d;
	GET DIAGNOSTICS affectedRow = ROW_COUNT;
    RETURN d;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION update_row(_t regclass, rowId UUID, d JSONB) RETURNS INT AS $$
declare
	affectedRow int;
BEGIN
    EXECUTE 'UPDATE '
        ||_t::regclass
        ||' SET data = $1 WHERE _id = $2'
        USING d, rowId;
	GET DIAGNOSTICS affectedRow = ROW_COUNT;
    RETURN affectedRow;
END;
$$
LANGUAGE plpgsql;

create extension if not exists plv8;

CREATE OR REPLACE FUNCTION query_by_id_and_populate(tbl text ,id uuid, to_populate json) RETURNS JSON AS $$
    // to_populate should contain:
	//			key: 'contacts.$._profile'          --> indicates the id path of the doc to be populated
	//                                                  if like '<key>.$.<key>' means to find the id in a object array
	//                                                  if like '<key>.<key>' means to find the id of a object
	//			from: 'my_schema.profile'           --> db name
	//			select: 'name email'                --> fields of the doc to be populated

    function set(obj, propString, value) {
        var propNames = propString.split('.'),
            tmpObj = obj;
            if (!obj) return obj;

        for (var i = 0; i < propNames.length; i++) {
            tmpObj = tmpObj[propNames[i]] = i !== propNames.length-1 ?  {} : value;
        }
        return obj;
    }

    function get(obj, propString) {
        var propNames = propString.split('.'),
        tmpObj = obj;

            for (var i = 0; i < propNames.length; i++) {
            if (! tmpObj ) return undefined;
                    tmpObj = tmpObj[propNames[i]];

        }
        return tmpObj;
    }

    function filter(obj, selectString) {
        var tmpObj = {};

        if (!selectString) return obj;

        var selectProp = selectString.split(' ');

        if (selectProp.length === 0) return obj;

        for (var i = 0; i < selectProp.length; i++ ){
            set(tmpObj, selectProp[i], get(obj, selectProp[i]));
        }

        return tmpObj;
    }

    var id_arr = [], doc, tmpId;

    // query_result is an array
    var query_result = plv8.execute('SELECT to_json(data) as data FROM ' + tbl + ' where _id = $1;', [id])[0];


    if ( query_result ) doc = query_result.data;

    // Now we get clean json document
    if ( !doc ) return undefined;

    if ( !to_populate || !to_populate.key || !to_populate.from) return doc;

    var key_array = to_populate.key.split('.$.');
    var key_of_doc = key_array[0], key_of_obj_array = key_array[1];

    if ( key_array.length === 1 ){      // Just populate sub object
        // populate sub object

        tmpId = get(doc, key_of_doc);
        if ( !tmpId) return doc;

        id_arr.push(tmpId);
    }
    else if ( key_array.length === 2){      // Populate object array
        // populate object array

        var obj_arr = get(doc, key_of_doc);

        if ( ! obj_arr || !(obj_arr instanceof Array)) return doc;

        for ( var i = 0; i < obj_arr.length; i++  ){
            tmpId  = get(obj_arr[i], key_of_obj_array);
            if ( tmpId ) id_arr.push(tmpId);
        }
    }
    else {
        // Dont know how to populate key patter with more than one '.$.'
        return doc;
    }

    if( id_arr.length === 0 ) return doc; // no id to populate

    query_result = plv8.execute('SELECT _id, to_json(data) as data FROM ' + to_populate.from_table + ' where _id = ANY(ARRAY(SELECT json_array_elements_text($1)::uuid));', [id_arr]);

    //return query_result;
    if ( query_result.length === 0 ) return doc;

    if ( key_array.length === 1){       // Just populate sub object
        set(doc, key_of_doc, filter(query_result[0].data, to_populate.select));
        return doc;
    }

    // Populate object array
    for ( var i = 0; i < query_result.length; i++  ){

        for ( var j = 0; j < obj_arr.length; j++ ){

            if (query_result[i]._id === get(obj_arr[j], key_of_obj_array)){

                set(obj_arr[j],key_of_obj_array, filter(query_result[i].data, to_populate.select));
                break;

            }
        }
    }

    return doc;
$$ LANGUAGE plv8 IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION json_merge_obj(data json, to_merge json) RETURNS JSON AS $$
    function merge(target, source) {
        if ( typeof target !== 'object' ) {
            target = {};
        }

        for (var property in source) {
            if ( source.hasOwnProperty(property) ) {
                var sourceProperty = source[ property ];

                if ( (typeof sourceProperty === 'object') && !(sourceProperty instanceof Array)) {
                    target[ property ] = merge( target[ property ], sourceProperty );
                    continue;
                }

                target[ property ] = sourceProperty;
            }
        }

        for (var a = 2, l = arguments.length; a < l; a++) {
            merge(target, arguments[a]);
        }
        return target;
    };

    if (data === undefined) return data;

    merge(data, to_merge);

    return data;
$$ LANGUAGE plv8 IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION json_update_array_property(data json, parameter json) RETURNS JSON AS $$
    function uniqueKey(o) {
        return Object.keys(o)[0];
    };

    function merge(target, source) {
        if ( typeof target !== 'object' ) {
            target = {};
        }

        for (var property in source) {
            if ( source.hasOwnProperty(property) ) {
                var sourceProperty = source[ property ];

                if ( (typeof sourceProperty === 'object') && !(sourceProperty instanceof Array)) {
                    target[ property ] = merge( target[ property ], sourceProperty );
                    continue;
                }

                target[ property ] = sourceProperty;
            }
        }

        for (var a = 2, l = arguments.length; a < l; a++) {
            merge(target, arguments[a]);
        }
        return target;
    };

    var op = uniqueKey(parameter);
    var arr_name = uniqueKey(parameter[op]);

    if( op === undefined || arr_name === undefined) {
        return data;
    }

    if ( /\$push/i.test(op) ){
        if ( data[arr_name] === undefined ){
            data[arr_name] = []
        }

        if ( data[arr_name] instanceof Array ){
            var to_be_pushed = parameter[op][arr_name];

            if (to_be_pushed instanceof Array ){
                Array.prototype.push.apply(data[arr_name], to_be_pushed);
            }else {
                data[arr_name].push(to_be_pushed);
            }


        }
    }
    else if ( /\$pull/i.test(op) ){
        if ( data[arr_name] === undefined || ! (data[arr_name] instanceof Array)){
            return data;
        }

        var whereKey = uniqueKey(parameter[op][arr_name]);
        var whereValue = parameter[op][arr_name][whereKey];

        if ( whereKey === undefined || whereValue === undefined ){
            return data;
        }

        for(var i = data[arr_name].length - 1; i >= 0;i--){
            if ( data[arr_name][i][whereKey] === whereValue ){
                data[arr_name].splice(i, 1);
            }
        }
    }
    else if ( /\$set/i.test(op) ){
        if ( data[arr_name] === undefined || ! (data[arr_name] instanceof Array)){
            return data;
        }

        var whereKey = uniqueKey(parameter[op][arr_name]['where']);
        var whereValue = parameter[op][arr_name]['where'][whereKey];
        var update = parameter[op][arr_name]['update'];

        if ( whereKey === undefined || whereValue === undefined || update === undefined){
            return data;
        }

        for(var i = data[arr_name].length - 1; i >= 0; i--){
            if ( data[arr_name][i][whereKey] === whereValue ){
                merge(data[arr_name][i], update);
            }
        }
    }

    return data;
$$ LANGUAGE plv8 IMMUTABLE STRICT;


-- This following function works.
--create types
DO $$
BEGIN
    IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'return_json') THEN
        CREATE TYPE return_json AS
        (
			_id uuid,
			data JSONB
        );
    END IF;
    --more types here...
END$$;

CREATE OR REPLACE FUNCTION select_by_id(_t text, rowId UUID) RETURNS setof return_json AS $$
declare
	_rec return_json;
BEGIN
	FOR _rec in
		EXECUTE
		'SELECT *
		FROM
		'||_t||'
		WHERE _id = $1'
--		LIMIT 1'
		USING rowId
	LOOP
		RETURN NEXT _rec;
	END LOOP;
END;
$$
LANGUAGE plpgsql;
